# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------
# Element 12: Displacement-based beam-column element

from G2Quadrature import *
from G2SectionWideFlange import *
from G2SectionRectangular import *

class element12():

    # -------------------------------------------------
    # Constructor
    # -------------------------------------------------
    def __init__(self, nsec, q, section, material, elno):

        self.type = "element12"
        self.q = q
        self.nsec = nsec

        # Create sections
        sections = []
        for i in range(self.nsec):
            if section[0] == 'WideFlange':
                hw = section[1]
                bf = section[2]
                tf = section[3]
                tw = section[4]
                nf = section[5]
                nw = section[6]
                sec = wfsection(hw, bf, tf, tw, nf, nw, material)
                sections.append(sec)
            elif section[0] == 'Rectangular':
                h = section[1]
                b = section[2]
                nfib = section[3]
                sec = rectangularSection(h, b, nfib, material)
                sections.append(sec)

        # Get integration parameters for an integration rule
        self.xip, self.weight = Gauss(self.nsec)

        # Set other instantiation variables
        self.no = elno
        self.secs = sections

        # Store basic forces to ease getResponse
        self.FbTrial = []
        self.FbCommitted = []

    # -------------------------------------------------
    # Set parameter
    # -------------------------------------------------
    def setParameter(self, parameter, value):

        if parameter == 'q':
            self.q = value
        else:
            for i in range(self.nsec):
                self.secs[i].setParameter(parameter, value)

    # -------------------------------------------------
    # Initialize
    # -------------------------------------------------
    def initialize(self, xyz):

        # Geometry restricted to 1,2 plane (x,y)
        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))

        # Check length
        if L < 1e-10:
            print('\n'"ERROR: Element number", self.no, "has zero length")
            import sys
            sys.exit()

    # -------------------------------------------------
    # State determination
    # -------------------------------------------------
    def state(self, xyz, ug, theLambda):

        # Element length
        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))

        # Direction cosines
        dx = dx / L

        # Transformation matrix
        Tbg = np.array([[-dx[1]/L, dx[0]/L, 1.0, dx[1]/L, -dx[0]/L, 0.0],
                      [-dx[1]/L, dx[0]/L, 0.0, dx[1]/L, -dx[0]/L, 1.0],
                      [-dx[0],  -dx[1],   0.0, dx[0],    dx[1],   0.0]])

        # Basic deformations
        ub = Tbg.dot(ug)

        # Integration points and weights
        points = self.xip
        weights = self.weight

        # Integrate basic force and basic stiffness matrix
        Fb = np.zeros(3)
        Kb = np.zeros((3, 3))
        for i in range(self.nsec):

            # Displacement interpolation evaluated at integration points
            x = (points[i] + 1.0) / 2.0
            Tsb1 = (6.0 * x - 4.0) / L
            Tsb2 = (6.0 * x - 2.0) / L
            Tsb3 = 1.0 / L

            # Get the section
            section = self.secs[i]

            # Transformation from Section to Basic
            Tsb = np.array([[Tsb1, Tsb2, 0.0],
                            [0.0,  0.0, Tsb3]])

            # Section deformations
            us = Tsb.dot(ub)

            # Section state determination
            Fs, Ks = section.state(us)

            # Sum section terms
            jacDet = L / 2
            Kb = Kb + np.einsum('ji,jk,kl->il', Tsb, Ks, Tsb) * (weights[i] * jacDet)
            Fb = Fb + np.transpose(Tsb).dot(Fs) * (weights[i] * jacDet)

        # Add element loads to the restoring forces
        q = theLambda * self.q
        mom = q * L**2 / 12.0
        barFb = [mom, -mom, 0.0]
        Fb = Fb + barFb

        # Store basic forces to avoid recomputing in getResponse
        self.FbTrial = Fb

        # Restoring force and stiffness matrix
        shear = q*L/2.0
        Fg = (np.transpose(Tbg)).dot(Fb) + np.array([-dx[1]*shear, dx[0]*shear, 0.0, -dx[1]*shear, dx[0]*shear, 0.0])
        Kg = np.einsum('ji,jk,kl->il', Tbg, Kb, Tbg)

        return Fg, Kg

    # -------------------------------------------------
    # DDM sensitivity analysis
    # -------------------------------------------------
    def stateDerivative(self, xyz, ug, theLambda, ddmParameter, ddmIndex, ddmIsHere, dKflag='none'):

        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))

        dx = dx / L

        Tbg = np.array([[-dx[1]/L, dx[0]/L, 1.0, dx[1]/L, -dx[0]/L, 0.0],
                        [-dx[1]/L, dx[0]/L, 0.0, dx[1]/L, -dx[0]/L, 1.0],
                        [-dx[0],  -dx[1],   0.0, dx[0],    dx[1],   0.0]])

        ub = Tbg.dot(ug)

        points = self.xip
        weights = self.weight

        dFb = np.zeros(3)
        dKb = np.zeros((3, 3))
        dKbdub = np.zeros((3,3,3))
        for ii in range(self.nsec):

            x = (points[ii] + 1.0) / 2.0
            Tsb1 = (6.0 * x - 4.0) / L
            Tsb2 = (6.0 * x - 2.0) / L
            Tsb3 = 1.0 / L

            section = self.secs[ii]

            Tsb = np.array([[Tsb1, Tsb2, 0.0],
                            [0.0,  0.0, Tsb3]])

            us = Tsb.dot(ub)

            dFs, dKs, dKsdus = section.stateDerivative(us, ddmParameter, ddmIndex, ddmIsHere, dKflag)

            dKb = dKb + (np.transpose(Tsb).dot(dKs)).dot(Tsb) * (weights[ii] * L / 2)
            dFb = dFb + np.transpose(Tsb).dot(dFs) * (weights[ii] * L / 2)

            dKbdub += (weights[ii] * L/2) * np.einsum('ki,knl,nm,lj->ijm', Tsb, dKsdus, Tsb, Tsb)

        dFg = (np.transpose(Tbg)).dot(dFb)
        dKg = np.einsum('ji,jk,kl->il', Tbg, dKb, Tbg)
        dKgdug = np.einsum('ki,knl,nm,lj->ijm', Tbg, dKbdub, Tbg, Tbg)

        if ddmParameter == 'q' and ddmIsHere:
            dmom = theLambda * L**2 / 12.0
            dFb_bar = [dmom, -dmom, 0.0]
            dshear = theLambda * L/2.0
            dFg = dFg + (np.transpose(Tbg)).dot(dFb_bar) + np.array([-dx[1]*dshear, dx[0]*dshear, 0.0, -dx[1]*dshear, dx[0]*dshear, 0.0])

        return dFg, dKg, dKgdug

    # -------------------------------------------------
    # Commit
    # -------------------------------------------------
    def commit(self, xyz, u):

        # Commit basic forces for getResponse
        self.FbCommitted = self.FbTrial

        # Commit section state
        for i in range(self.nsec):
            section = self.secs[i]
            section.commit()


    # -------------------------------------------------
    # Commit sensitivity history variables
    # -------------------------------------------------
    def commitSensitivity(self, xyz, ug, ddmug, ddmParameter, ddmIndex, ddmIsHere):

        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))
        dx = dx / L
        Tbg = np.array([[-dx[1]/L, dx[0]/L, 1.0, dx[1]/L, -dx[0]/L, 0.0],
                      [-dx[1]/L, dx[0]/L, 0.0, dx[1]/L, -dx[0]/L, 1.0],
                      [-dx[0],  -dx[1],   0.0, dx[0],    dx[1],   0.0]])
        ub = Tbg.dot(ug)
        ddmub = Tbg.dot(ddmug)
        points = self.xip
        for i in range(self.nsec):
            x = (points[i] + 1.0) / 2.0
            Tsb1 = (6.0 * x - 4.0) / L
            Tsb2 = (6.0 * x - 2.0) / L
            Tsb3 = 1.0 / L
            section = self.secs[i]
            Tsb = np.array([[Tsb1, Tsb2, 0.0],
                            [0.0,  0.0, Tsb3]])
            us = Tsb.dot(ub)
            ddmus = Tsb.dot(ddmub)

            # For the ASCE-JSE paper
            error1 = False
            if not error1:
                section.commitSensitivity(us, ddmus, ddmParameter, ddmIndex, ddmIsHere)


    # -------------------------------------------------
    # Print response
    # -------------------------------------------------
    def getResponse(self, xyz):

        # Geometry
        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))

        # Print element forces
        shear = (self.FbCommitted[0] + self.FbCommitted[1]) / L
        [N1, N2] = self.FbCommitted[2] * np.array([-1, 1])
        [V1, V2] = shear * np.array([1, - 1])
        [M1, M2] = [self.FbCommitted[0], self.FbCommitted[1]]

        print("Element %d (%s): N1=%.1f / N2=%.1f / V1=%.1f / V2=%.1f / M1=%.1f / M2=%.1f" % (self.no, self.type, N1, N2, V1, V2, M1, M2))
