# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------
# Element 3: Truss element with "moderate" displacements

import numpy as np

class element3():

    # -------------------------------------------------
    # Constructor
    # -------------------------------------------------
    def __init__(self, E, A, N0, elno):

        self.type = "element3"

        self.no = elno
        self.E = E
        self.A = A
        self.N0 = N0

        # Store basic forces to ease getResponse
        self.sTrial = []
        self.sCommitted = []

    # -------------------------------------------------
    # Initialize
    # -------------------------------------------------
    def initialize(self, xyz):

        # Geometry restricted to 1,2 plane (x,y)
        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))

        # Check length
        if L < 1e-10:
            print('\n'"ERROR: Element number", self.no, "has zero length")
            import sys
            sys.exit()

    # -------------------------------------------------
    # State determination
    # -------------------------------------------------
    def state(self, xyz, U, theLambda):

        # Reminder:
        # The first column of U are the current total displacements
        # The second column of U are the displacement since last committed state
        # Third column of U, not used here, is the solution to the system of equations at this iteration

        # Original element length
        dx0 = xyz[1,:] - xyz[0,:]
        L0 = np.sqrt(dx0.dot(dx0))

        # Current element length
        du = U[[2, 3], 0] - U[[0, 1], 0]
        dx = dx0 + du
        L = np.sqrt(dx.dot(dx))

        # Green-Lagrange strain
        es = 0.5 * (L**2 / L0**2 - 1)

        # Force and stiffness in the basic system
        kb = self.E * self.A / L
        self.sTrial = self.E * self.A * es + self.N0

        # Previously committed geometry
        duStep = U[[2, 3], 1] - U[[0, 1], 1]
        dxCommitted = dx0 + du - duStep
        LCommitted = np.sqrt(dxCommitted.dot(dxCommitted))

        # Calculate "direction cosines"
        dxCommitted = dxCommitted / LCommitted
        duStep = duStep / L

        # Rotation of Local member since last commit (notice that displacements above are in Global system)
        theta = (duStep).dot([-dxCommitted[1], dxCommitted[0]])

        # Transformation from Basic to Local
        Tbl = np.array([-1, -theta, 1, theta])

        # Element stiffness and force in local coordinate system
        Kl = np.outer(Tbl.dot(kb), Tbl)
        tildeFl = Tbl.dot(self.sTrial)

        # Add geometric stiffness matrix
        c = np.array([0, -1, 0, 1])
        kGeometric = np.outer(c, c) * self.sTrial / L
        Kl += kGeometric

        # Transform from Local to Global
        Tlg = np.array([[dxCommitted[0], dxCommitted[1], 0.0, 0.0],
                      [-dxCommitted[1],  dxCommitted[0], 0.0, 0.0],
                      [0.0, 0.0,         dxCommitted[0], dxCommitted[1]],
                      [0.0, 0.0,        -dxCommitted[1], dxCommitted[0]]])
        tildeFg = (np.transpose(Tlg)).dot(tildeFl)
        Kg = ((np.transpose(Tlg)).dot(Kl)).dot(Tlg)

        return tildeFg, Kg

    # -------------------------------------------------
    # Commit
    # -------------------------------------------------
    def commit(self, xyz, u):

        # Commit basic forces for getResponse
        self.sCommitted = self.sTrial

    # -------------------------------------------------
    # Print response
    # -------------------------------------------------
    def getResponse(self, xyz):

        # Print element force
        N = self.sCommitted
        print("Element %d (%s): N=%.1f " % (self.no, self.type, N))
