# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------
# Element 5: Linear elastic frame element

import numpy as np

class element5():

    # -------------------------------------------------
    # Constructor
    # -------------------------------------------------
    def __init__(self, E, A, I, q, elno):

        self.type = "element5"

        # Store input
        self.no = elno
        self.E  = E
        self.A  = A
        self.I  = I
        self.q  = q

        # Store basic forces to ease getResponse
        self.FbTrial = []
        self.FbCommitted = []

    # -------------------------------------------------
    # Set parameter
    # -------------------------------------------------
    def setParameter(self, parameter, value):

        if parameter == 'E':
            self.E = value
        elif parameter == 'A':
            self.A = value
        elif parameter == 'I':
            self.I = value
        elif parameter == 'q':
            self.q = value
        else:
            print("Cannot set parameter", parameter, "in element", self.no)

    # -------------------------------------------------
    # Initialize
    # -------------------------------------------------
    def initialize(self, xyz):

        # Geometry restricted to 1,2 plane (x,y)
        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))

        # Check length
        if L < 1e-10:
            print('\n'"ERROR: Element number", self.no, "has zero length")
            import sys
            sys.exit()

    # -------------------------------------------------
    # State determination
    # -------------------------------------------------
    def state(self, xyz, ugMatrix, theLambda):

        # Total displacements
        ug = ugMatrix[:, 0]

        # Geometry
        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))
        dx = dx / L

        # Transformation matrix from basic to global
        Tbg = np.array([[-dx[1]/L, dx[0]/L, 1.0, dx[1]/L, -dx[0]/L, 0.0],
                        [-dx[1]/L, dx[0]/L, 0.0, dx[1]/L, -dx[0]/L, 1.0],
                        [-dx[0],  -dx[1],   0.0, dx[0],    dx[1],   0.0]])

        # Form basic stiffness matrix
        EI = self.E * self.I
        EA = self.E * self.A
        Kb = np.array([[4*EI/L, 2*EI/L, 0.0],
                       [2*EI/L, 4*EI/L, 0.0],
                       [0.0,    0.0,    EA/L]])

        # Initial basic forces (clamping forces)
        q = theLambda * self.q
        mom = q * L**2 / 12.0
        Fb_bar = [mom, -mom, 0.0]

        # Basic force
        Fb = (Kb.dot(Tbg)).dot(ug) + Fb_bar

        # Store basic forces to avoid recomputing in getResponse
        self.FbTrial = Fb

        # Restoring force and stiffness matrix in global element configuration
        shear = q*L/2.0
        Fg_tilde = (np.transpose(Tbg)).dot(Fb) + np.array([-dx[1]*shear, dx[0]*shear, 0.0, -dx[1]*shear, dx[0]*shear, 0.0])
        Kg = (np.transpose(Tbg).dot(Kb)).dot(Tbg)

        return Fg_tilde, Kg

    # -------------------------------------------------
    # DDM sensitivity analysis
    # -------------------------------------------------
    def stateDerivative(self, xyz, ugMatrix, theLambda, ddmParameter, ddmIndex=0, ddmIsHere=False, dkflag='none'):

        ug = ugMatrix[:, 0]

        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))
        dx = dx / L

        Tbg = np.array([[-dx[1]/L, dx[0]/L, 1.0, dx[1]/L, -dx[0]/L, 0.0],
                        [-dx[1]/L, dx[0]/L, 0.0, dx[1]/L, -dx[0]/L, 1.0],
                        [-dx[0],  -dx[1],   0.0, dx[0],    dx[1],   0.0]])

        if ddmParameter == 'E' and ddmIsHere:
            dEI = self.I
            dEA = self.A
        elif ddmParameter == 'A' and ddmIsHere:
            dEI = 0.0
            dEA = self.E
        elif ddmParameter == 'I' and ddmIsHere:
            dEI = self.E
            dEA = 0.0
        else:
            dEI = 0.0
            dEA = 0.0

        dKb = np.array([[4*dEI/L, 2*dEI/L, 0.0],
                        [2*dEI/L, 4*dEI/L, 0.0],
                        [0.0,    0.0,    dEA/L]])

        dFb = (dKb.dot(Tbg)).dot(ug)
        dFg = (np.transpose(Tbg)).dot(dFb)
        dKg = (np.transpose(Tbg).dot(dKb)).dot(Tbg)

        if ddmParameter == 'q':
            q = theLambda
            mom = q * L**2 / 12.0
            dFb_bar = [mom, -mom, 0.0]
            shear = q*L/2.0
            dFg = (np.transpose(Tbg)).dot(dFb_bar) + np.array([-dx[1]*shear, dx[0]*shear, 0.0, -dx[1]*shear, dx[0]*shear, 0.0])

        dKgdug = 0.0

        return dFg, dKg, dKgdug

    # -------------------------------------------------
    # Commit
    # -------------------------------------------------
    def commit(self, xyz, ug):

        # Commit basic forces for getResponse
        self.FbCommitted = self.FbTrial


    # -------------------------------------------------
    # Commit sensitivity history variables
    # -------------------------------------------------
    def commitSensitivity(self, xyz, ug, ddmug, ddmParameter, ddmIndex, ddmIsHere):

        # Nothing to do here for a linear element
        return

    # -------------------------------------------------
    # Print response
    # -------------------------------------------------
    def getResponse(self, xyz):

        # Geometry
        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))

        # Print element forces
        shear = (self.FbCommitted[0] + self.FbCommitted[1]) / L
        shear0 = self.q * L / 2.0


        [N1, N2] = self.FbCommitted[2] * np.array([-1, 1])
        [V1, V2] = shear * np.array([1, - 1]) + shear0 * np.array([1, 1])
        [M1, M2] = [self.FbCommitted[0], self.FbCommitted[1]]

        print("Element %d (%s): N1=%.1f / N2=%.1f / V1=%.1f / V2=%.1f / M1=%.1f / M2=%.1f" % (self.no, self.type, N1, N2, V1, V2, M1, M2))
