# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------
# Element 6: Linear elastic frame element with P-delta

import numpy as np

class element6():

    # -------------------------------------------------
    # Constructor
    # -------------------------------------------------
    def __init__(self, option, E, A, I, q, elno):

        self.type = "element6"

        # Store input
        self.option = option
        self.no = elno
        self.E  = E
        self.A  = A
        self.I  = I
        self.q  = q

        # Store basic forces to ease getResponse
        self.FbTrial = []
        self.FbCommitted = []

    # -------------------------------------------------
    # Set parameter
    # -------------------------------------------------
    def setParameter(self, parameter, value):

        if parameter == 'E':
            self.E = value
        elif parameter == 'A':
            self.A = value
        elif parameter == 'I':
            self.I = value
        elif parameter == 'q':
            self.q = value
        else:
            print("Cannot set parameter", parameter, "in element", self.no)

    # -------------------------------------------------
    # Initialize
    # -------------------------------------------------
    def initialize(self, xyz):

        # Geometry restricted to 1,2 plane (x,y)
        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))

        # Check length
        if L < 1e-10:
            print('\n'"ERROR: Element number", self.no, "has zero length")
            import sys
            sys.exit()

    # -------------------------------------------------
    # State determination
    # -------------------------------------------------
    def state(self, xyz, ugMatrix, theLambda):

        # Total displacements
        ug = ugMatrix[:, 0]

        # Geometry
        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))
        cx = dx[0]/L
        cy = dx[1]/L

        # Basic elastic stiffness matrix
        EI = self.E * self.I
        EA = self.E * self.A
        Kb = np.array([[4*EI/L, 2*EI/L, 0.0],
                       [2*EI/L, 4*EI/L, 0.0],
                       [0.0,    0.0,    EA/L]])

        # Cannot include Big P-delta in Basic configuration, so here go via Local
        # Note on G2: Counterclockwise rotations! Axial DOF is LAST in Basic configuration!
        Tbl = np.array([[0.0,  1/L, 1.0, 0.0, -1/L, 0.0],
                        [0.0,  1/L, 0.0, 0.0, -1/L, 1.0],
                        [-1.0, 0.0, 0.0, 1.0,  0.0, 0.0]])

        # Transformation from local to global
        Tlg = np.array([[ cx, cy,  0.0, 0.0, 0.0, 0.0],
                        [-cy, cx,  0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0,  cx,  cy, 0.0],
                        [0.0, 0.0, 0.0, -cy,  cx, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0, 1.0]])

        # Local deformations
        ul = Tlg.dot(ug)

        # Local elastic stiffness matrix
        Kl = (np.transpose(Tbl).dot(Kb)).dot(Tbl)

        # Basic clamping forces
        q = theLambda * self.q
        mom = q * L**2 / 12.0
        Fb_bar = [mom, -mom, 0.0]

        # Local elastic forces, including member load
        shear = q*L/2.0
        Fl = Kl.dot(ul)
        Fl += (np.transpose(Tbl)).dot(Fb_bar)
        Fl += np.array([0.0, shear, 0.0, 0.0, shear, 0.0])

        # Extract axial force (compression positive here)
        P = Fl[0]

        # Geometric stiffness matrix (using DOF number from website notes, not G2)
        if self.option == "None":
            Klgeom = np.zeros((6, 6))
        elif self.option == "Big":
            Klgeom = np.array([[0.0,  0.0, 0.0, 0.0,  0.0, 0.0],
                               [0.0,  1/L, 0.0, 0.0, -1/L, 0.0],
                               [0.0,  0.0, 0.0, 0.0,  0.0, 0.0],
                               [0.0,  0.0, 0.0, 0.0,  0.0, 0.0],
                               [0.0, -1/L, 0.0, 0.0,  1/L, 0.0],
                               [0.0,  0.0, 0.0, 0.0,  0.0, 0.0]])

        elif self.option == "BigSmall":
            Klgeom = np.array([[0.0,  0.0,     0.0,    0.0, 0.0,       0.0],
                               [0.0,  6/(5*L), -1/10,  0.0, -6/(5*L), -1/10],
                               [0.0, -1/10,    2*L/15, 0.0, 1/10,     -L/30],
                               [0.0,  0.0,     0.0,    0.0, 0.0,       0.0],
                               [0.0, -6/(5*L), 1/10,   0.0, 6/(5*L),   1/10],
                               [0.0, -1/10,    -L/30,  0.0, 1/10,      2*L/15]])
        else:
            print("Error: Option given to Element 6 is not yet implemented.")
            import sys
            sys.exit()

        # Signs that flip when going to G2 conventions (counterclockwise rotation DOFs)
        Flip = np.ones((6, 6))
        Flip[2, 1] = -1
        Flip[1, 2] = -1
        Flip[5, 1] = -1
        Flip[1, 5] = -1
        Flip[4, 2] = -1
        Flip[2, 4] = -1
        Flip[5, 4] = -1
        Flip[4, 5] = -1

        # Update signs
        Klgeom = np.multiply(Klgeom, Flip)

        # Add local forces due to the change in geometry
        Fl -= P * Klgeom.dot(ul)

        # Restoring force and stiffness matrix in global element configuration
        Fg_tilde = (np.transpose(Tlg)).dot(Fl)
        Kg = (np.transpose(Tlg).dot(Kl-Klgeom)).dot(Tlg)

        return Fg_tilde, Kg


    # -------------------------------------------------
    # Commit
    # -------------------------------------------------
    def commit(self, xyz, ug):

        return

    # -------------------------------------------------
    # Print response
    # -------------------------------------------------
    def getResponse(self, xyz):

        # Print element force
        print("Element %d (%s): No information " % (self.no, self.type))

        return
