# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------
# Element 9: One-component series system (with uniaxial materials)

import numpy as np
from G2MaterialBilinear import *

class element9():

    # -------------------------------------------------
    # Constructor
    # -------------------------------------------------
    def __init__(self, section, material, elno):

        self.type = "element9"

        # Store input
        self.no = elno
        self.EA = float(section[1])
        self.EI = float(section[2])

        # Create materials
        if material[0] == 'Bilinear':
            self.spring1 = bilinearMaterial(material)
            self.spring2 = bilinearMaterial(material)
        else:
            print("Error: Wrong material given to element type 9")
            import sys
            sys.exit()

        # Store basic forces to ease getResponse
        self.FbTrial = []
        self.FbCommitted = np.zeros(3)


# -------------------------------------------------
    # Initialize
    # -------------------------------------------------
    def initialize(self, xyz):

        # Geometry restricted to 1,2 plane (x,y)
        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))

        # Check length
        if L < 1e-10:
            print('\n'"ERROR: Element number", self.no, "has zero length")
            import sys
            sys.exit()

    # -------------------------------------------------
    # State determination
    # -------------------------------------------------
    def state(self, xyz, ug, theLambda):

        # Pick up relevant deformations
        ug = ug[:, 1]

        # Direction cosines
        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))
        dx = dx / L

        # Transformation matrix from basic to global
        Tbg = np.array([[-dx[1]/L, dx[0]/L, 1.0, dx[1]/L, -dx[0]/L, 0.0],
                        [-dx[1]/L, dx[0]/L, 0.0, dx[1]/L, -dx[0]/L, 1.0],
                        [-dx[0],  -dx[1],   0.0, dx[0],    dx[1],   0.0]])

        # Basic deformations (rot-left, rot-right, elongation)
        ub = Tbg.dot(ug)

        # Similar to Element 13, compute the Basic forces that produce section deformations compatible with ub
        Fb, Kb = self.compatibility(ub, L)

        # Store trial basic forces
        self.FbTrial = Fb

        # Global restoring force and stiffness matrix
        Fg = (np.transpose(Tbg)).dot(Fb)
        Kg = np.einsum('ji,jk,kl->il', Tbg, Kb, Tbg)

        return Fg, Kg

    # -------------------------------------------------
    # Compatibility iterations
    # -------------------------------------------------
    def compatibility(self, ub, L):

        # Cross-section stiffness values
        EAL = self.EA / L
        EIL = self.EI / L

        # Flexibility matrix
        fbElastic = np.array([[1.0/(3*EIL), -1.0/(6*EIL), 0.0],
                              [-1.0/(6*EIL), 1.0/(3*EIL), 0.0],
                              [0.0,          0.0,         1.0/EAL]])

        # Set iteration controls
        maxiter = 10
        tol = 1e-6

        # Initialize vector of incremental forces
        FbIncrement = np.zeros(3)

        # Compatibility iterations to determine the Fb that gives (ubTrial-ub)=0
        for j in range(maxiter):

            # Residual deformations
            ubResidual = np.copy(ub)

            # Hinge forces from Basic forces (Fs = Ttilde * Fb)
            Fspring1Trial = self.FbCommitted[0] + FbIncrement[0]
            Fspring2Trial = self.FbCommitted[1] + FbIncrement[1]

            # Equilibrium iterations to determine the uHinges that gives (FspringsTrial-Fsprings)=0
            uHinge1, Kspring1 = self.spring1.equilibrium(Fspring1Trial)
            uHinge2, Kspring2 = self.spring2.equilibrium(Fspring2Trial)

            # Reduction in residual due to deformation in springs
            ubResidual[0] -= uHinge1
            ubResidual[1] -= uHinge2

            # Reduction in due to deformation of the elastic interior (only from force increment)
            ubResidual -= np.dot(fbElastic, FbIncrement)

            # Flexibility matrix
            fb = np.copy(fbElastic)
            fb[0, 0] += 1/Kspring1
            fb[1, 1] += 1/Kspring2

            # Compute next trial for the basic force vector, if we didn't converge
            FbIncrement = np.linalg.solve(fb, ubResidual) + FbIncrement

            # Check convergence
            if np.linalg.norm(ubResidual) < tol:
                break

        if np.linalg.norm(ubResidual) >= tol:
            print("ERROR: Compatibility iterations in element", self.no, "failed")
            import sys
            sys.exit()

        Fb = self.FbCommitted + FbIncrement
        Kb = np.linalg.inv(fb)

        return Fb, Kb


    # -------------------------------------------------
    # Commit
    # -------------------------------------------------
    def commit(self, xyz, u):

        self.FbCommitted = self.FbTrial
        self.spring1.commit()
        self.spring2.commit()


# -------------------------------------------------
    # Print response
    # -------------------------------------------------
    def getResponse(self, xyz):

        # Print element force
        print("Element %d (%s): No information " % (self.no, self.type))
